import uuid
import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from notihub.notifiers.aws.clients.base_aws_client import BaseAWSClient


@dataclass
class PinpointAddress:
    """A recipient address for Pinpoint, including token and service."""

    token: str
    service: str


@dataclass
class PinpointClient(BaseAWSClient):
    """
    PinpointClient

    Class used to generate notifications via AWS Pinpoint
    """
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str

    def __post_init__(self):
        self.pinpoint_client = self.initialize_client("pinpoint")

    def _prepare_custom_data(
        self,
        custom_data: Optional[Dict[str, Any]],
        *,
        deep_link_url: Optional[str],
        image_url: Optional[str],
    ) -> Dict[str, Any]:
        """Return a copy of custom_data enriched with deeplink / image_url."""
        data: Dict[str, Any] = (custom_data or {}).copy()
        data.update({"deeplink": deep_link_url, "image_url": image_url})
        return data

    def _build_apns_message(
        self,
        *,
        title: str,
        body: str,
        action: str,
        deep_link_url: Optional[str],
        image_url: Optional[str],
        silent_push: bool,
        custom_data: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Build the APNS message."""
        aps_payload: Dict[str, Any] = {"sound": "default"}

        if silent_push:
            aps_payload["content-available"] = 1
        else:
            aps_payload["alert"] = {"title": title, "body": body}

        if image_url:
            aps_payload["mutable-content"] = 1

        payload_root = {"aps": aps_payload, **custom_data}
        if deep_link_url:
            payload_root["Url"] = deep_link_url

        message: Dict[str, Any] = {
            "Action": action,
            "RawContent": json.dumps(payload_root),
        }
        if deep_link_url:
            message["Url"] = deep_link_url

        return message

    def _build_gcm_message(
        self,
        *,
        title: str,
        body: str,
        action: str,
        deep_link_url: Optional[str],
        image_url: Optional[str],
        silent_push: bool,
        custom_data: Dict[str, Any],
        time_to_live: Optional[int],
        priority: Optional[str],
    ) -> Dict[str, Any]:
        payload_data = {**custom_data, "title": title, "body": body, "Url": deep_link_url}

        gcm_payload: Dict[str, Any] = {"data": payload_data}
        if silent_push:
            gcm_payload["content_available"] = 1
        else:
            notification = {"title": title, "body": body}
            if image_url:
                notification["image"] = image_url

            gcm_payload["notification"] = notification

        if priority:
            gcm_payload["priority"] = priority

        if time_to_live is not None:
            gcm_payload["time_to_live"] = time_to_live

        message: Dict[str, Any] = {
            "Action": action,
            "RawContent": json.dumps(gcm_payload),
        }
        if deep_link_url:
            message["Url"] = deep_link_url

        return message

    def send_pinpoint_push_notification(
        self,
        *,
        application_id: str,
        addresses: List[PinpointAddress],
        title: str,
        body: str,
        deep_link_url: Optional[str] = None,
        image_url: Optional[str] = None,
        custom_data: Optional[Dict[str, Any]] = None,
        silent_push: bool = False,
        time_to_live: Optional[int] = None,
        priority: Optional[str] = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        Sends a push notification via Pinpoint to specified addresses.
        It groups addresses by service type (GCM, APNS) and sends a batch for each.
        """
        action = "DEEP_LINK" if deep_link_url else "OPEN_APP"
        processed_custom_data = self._prepare_custom_data(
            custom_data, deep_link_url=deep_link_url, image_url=image_url
        )
        addresses_by_service: Dict[str, List[str]] = {}
        for address in addresses:
            service = address.service.upper()
            if service not in addresses_by_service:
                addresses_by_service[service] = []

            addresses_by_service[service].append(address.token)

        responses = []
        for service, tokens in addresses_by_service.items():
            message_config: Dict[str, Any] = {}
            if service == "GCM":
                message_config["GCMMessage"] = self._build_gcm_message(
                    title=title,
                    body=body,
                    action=action,
                    deep_link_url=deep_link_url,
                    image_url=image_url,
                    silent_push=silent_push,
                    custom_data=processed_custom_data,
                    time_to_live=time_to_live,
                    priority=priority,
                )
            elif service in ["APNS", "APNS_SANDBOX"]:
                message_config["APNSMessage"] = self._build_apns_message(
                    title=title,
                    body=body,
                    action=action,
                    deep_link_url=deep_link_url,
                    image_url=image_url,
                    silent_push=silent_push,
                    custom_data=processed_custom_data,
                )
            else:
                continue

            message_request = {
                "Addresses": {token: {"ChannelType": service} for token in tokens},
                "MessageConfiguration": message_config,
            }
            response = self.pinpoint_client.send_messages(
                ApplicationId=application_id, MessageRequest=message_request
            )
            responses.append(response)

        return responses

    def create_pinpoint_endpoint(
        self,
        *,
        application_id: str,
        device_token: str,
        channel_type: str,
        user_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Creates a new Pinpoint endpoint with a generated unique ID.
        Args:
            application_id: The Pinpoint Application ID.
            device_token: The push notification token from APNS/FCM.
            channel_type: Channel type (e.g., "GCM", "APNS", "APNS_SANDBOX").
            user_id: The user ID of the application.
        Returns:
            Dict containing API response and the generated 'EndpointId'.
            Store the 'EndpointId' for future updates/deletions.
        Raises:
            ClientError: If the Pinpoint API call fails.
        """
        generated_endpoint_id = str(uuid.uuid4())
        endpoint_request: Dict[str, Any] = {
            "Address": device_token,
            "ChannelType": channel_type.upper(),
            "OptOut": "NONE",
        }
        if user_id:
            endpoint_request["UserId"] = user_id

        response = self.pinpoint_client.update_endpoint(
            ApplicationId=application_id,
            EndpointId=generated_endpoint_id,
            EndpointRequest=endpoint_request,
        )
        return {"EndpointId": generated_endpoint_id, **response}

    def update_pinpoint_endpoint(
        self,
        *,
        application_id: str,
        endpoint_id: str,
        device_token: Optional[str] = None,
        channel_type: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Updates specific attributes of an existing Pinpoint endpoint.
        Args:
            application_id: The Pinpoint Application ID.
            endpoint_id: The unique ID of the endpoint to update.
            device_token: Optional new push notification token.
            channel_type: Optional channel type.
            user_id: Optional application's user ID.
        Returns:
            The response from the Pinpoint update_endpoint operation.
        Raises:
            ClientError: If the Pinpoint API call fails.
        """
        endpoint_request: Dict[str, Any] = {}

        if device_token:
            endpoint_request["Address"] = device_token

        if channel_type:
            endpoint_request["ChannelType"] = channel_type.upper()

        if user_id:
            endpoint_request["UserId"] = user_id

        response = self.pinpoint_client.update_endpoint(
            ApplicationId=application_id,
            EndpointId=endpoint_id,
            EndpointRequest=endpoint_request,
        )
        return response

    def delete_pinpoint_endpoint(
        self,
        *,
        application_id: str,
        endpoint_id: str,
    ) -> Dict[str, Any]:
        """Deletes a specific Pinpoint endpoint.
        Args:
            application_id: The Pinpoint Application ID.
            endpoint_id: The unique identifier of the endpoint to delete.
        Returns:
            The response from the Pinpoint delete_endpoint operation.
        Raises:
            ClientError: If the Pinpoint API call fails.
        """
        response = self.pinpoint_client.delete_endpoint(
            ApplicationId=application_id,
            EndpointId=endpoint_id,
        )
        return response



from dataclasses import dataclass
from abc import ABC

import boto3

@dataclass
class BaseAWSClient(ABC):
    
    aws_access_key_id: str = None
    aws_secret_access_key: str = None
    region_name: str = None

    def initialize_client(self, service_name: str = None):
        """
        Initialize the AWS client
        
        Args:
            service_name (str): The name of the service
        """
        self.validate_credentials()
        kwargs = {}
        if self.aws_access_key_id and self.aws_secret_access_key and self.region_name:
            kwargs["aws_access_key_id"] = self.aws_access_key_id
            kwargs["aws_secret_access_key"] = self.aws_secret_access_key
            kwargs["region_name"] = self.region_name

        return boto3.client(service_name, **kwargs)

    def validate_credentials(self):
        """
        Validate AWS credentials are provided
        """
        required_credentials = list(
            filter(None, [self.aws_access_key_id, self.aws_secret_access_key, self.region_name])
        )
        if required_credentials and not len(required_credentials) == 3:
            raise ValueError(
                "AWS access key ID, secret access key, "
                "and region name must be provided all together."
            )




"""

Twilio Notifier

This module contains the Twilio notifier class which is used to send notifications via Twilio
"""

from dataclasses import dataclass
from typing import List, Optional

from twilio.rest import Client

from notihub.base_notifier import BaseNotifier


@dataclass
class TwilioNotifier(BaseNotifier):
    """
    TwilioNotifier

    Class used to send notifications via Twilio
    """

    account_sid: Optional[str] = None
    auth_token: Optional[str] = None
    twilio_phone_number: Optional[str] = None

    def __post_init__(self):
        if not self.account_sid or not self.auth_token:
            raise ValueError(
                "Twilio Account SID and Auth Token must be provided."
            )

        self.client = Client(self.account_sid, self.auth_token)

    def send_sms_notification(self, phone_number: str, message: str, **kwargs) -> str:
        """Sends a SMS notification to the given phone number using Twilio."""
        if not self.twilio_phone_number:
            raise ValueError("Twilio phone number must be provided for sending SMS.")

        message = self.client.messages.create(
            to=phone_number,
            from_=self.twilio_phone_number,
            body=message,
            **kwargs,
        )
        return message.sid

    def send_email_notification(
        self,
        *,
        subject: str,
        email_data: dict,
        recipients: List[str],
        sender: str,
        template: str,
        cc_emails: List[str] = None,
        bcc_emails: List[str] = None,
        **kwargs,
    ) -> str:
        """Sends an email notification to the given email (not supported by Twilio directly for this library)."""
        raise NotImplementedError("Twilio does not directly support email notifications in this library.")

    def send_push_notification(self, device: str, message: str, **kwargs) -> str:
        """Sends a push notification to the given message (not supported by Twilio directly for this library)."""
        raise NotImplementedError("Twilio does not directly support push notifications in this library.")



import json
from dataclasses import dataclass
from typing import Any, Dict, Union

from notihub.notifiers.aws.clients.base_aws_client import BaseAWSClient



@dataclass
class SNSClient(BaseAWSClient):
    """
    SNSClient

    Class used to generate notifications via AWS SNS
    """

    def __post_init__(self):
        self.sns_client = self.initialize_client("sns")

    def get_topic(self, topic_arn: str) -> Dict[str, Any]:
        """
        Gets a topic with the given ARN

        Args:
            topic_arn (str): The ARN of the topic

        Returns:
            dict: Response of the client operation
        """
        return self.sns_client.get_topic_attributes(TopicArn=topic_arn)

    def delete_topic(self, topic_arn: str) -> Dict[str, Any]:
        """
        Deletes a topic with the given ARN

        Args:
            topic_arn (str): The ARN of the topic

        Returns:
            dict: Response of the client operation
        """
        return self.sns_client.delete_topic(TopicArn=topic_arn)

    def create_topic(self, topic_name: str) -> Dict[str, Any]:
        """
        Creates a topic with the given name

        Args:
            topic_name (str): The name of the topic

        Returns:
            dict: response of the client operation with the ARN of the topic
        """
        return self.sns_client.create_topic(
            Name=topic_name,
        )

    def subscribe_to_topic(
        self, topic_arn: str, protocol: str, endpoint: str
    ) -> Dict[str, Any]:
        """
        Subscribes the given endpoint to the given topic

        Args:
            topic_arn (str): The topic ARN
            protocol (str): The protocol to use
            endpoint (str): The endpoint to subscribe to

        Returns:
            dict: response of the client operation with the ARN of the subscription
        """
        return self.sns_client.subscribe(
            TopicArn=topic_arn,
            Protocol=protocol,
            Endpoint=endpoint,
        )

    def send_topic_notification(
        self,
        *,
        topic_arn: str,
        message: str,
        subject: str,
        message_structure: str = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """

        Sends a notification to the given topic

        Args:
            topic_arn (str): The topic ARN
            message (str): The message to send

        Additional arguments:
            subject (str): The subject of the message
            target_arn (str): The target ARN
            message_structure (str): The message structure
            \*args: Additional arguments
            \*\*kwargs: Additional keyword arguments

        Returns:
            dict: Response of the client operation
        """
        return self.sns_client.publish(
            TopicArn=topic_arn,
            Message=message,
            Subject=subject,
            MessageStructure=message_structure or "",
            **kwargs,
        )

    def send_sms_notification(
        self, phone_number: str, message: str, **kwargs
    ) -> Dict[str, Any]:
        """
        Sends a SMS notification to the given phone number

        Args:
            phone_number (str): The phone number to send the message to
            message (str): The message to send

        Additional arguments:
            \*\*kwargs: Additional keyword arguments

        Returns:
            dict: Response of the client operation
        """
        return self.sns_client.publish(
            PhoneNumber=phone_number,
            Message=message,
            **kwargs,
        )

    def send_push_notification(
        self,
        device: str,
        message: str,
        title: str,
        payload: Union[Dict[str, Any], None] = None,
        **kwargs,
    ) -> str:
        """
        Sends a push notification with a title to the given message

        Args:
            device (str): The device to send the message to
            title (str): The title of the push notification
            message (str): The message to send
            payload (dict, optional): Custom payload to send. If not provided,
            a default will be used.

        Returns:
            dict: Response of the client operation
        """
        if payload is None:
            payload = {
                "default": message,
                "APNS": json.dumps({"aps": {"alert": {"title": title, "body": message}}}),
                "APNS_SANDBOX": json.dumps(
                    {"aps": {"alert": {"title": title, "body": message}}}
                ),
                "GCM": json.dumps({"notification": {"title": title, "body": message}}),
            }

        return self.sns_client.publish(
            TargetArn=device, Message=json.dumps(payload), MessageStructure="json"
        )

    def create_device_endpoint(
        self,
        platform_application_arn: str,
        device_token: str,
        custom_user_data: str = "",
        **kwargs,
    ):
        """
        Creates a platform endpoint for the given device token.

        Args:
            platform_application_arn (str): The ARN of the platform application
            (e.g., APNS).
            device_token (str): The token associated with the device to register.
            custom_user_data (str, optional): The custom user data to associate with the
            device endpoint. This should be a JSON-formatted string
            representing user-specific data, such as user ID, subscription type, etc.
            If not provided, no custom user data is associated with the endpoint.
            Defaults to "".

        Returns:
            dict: Response from the SNS client operation,
                which includes the platform endpoint details
                or an error message if the operation fails.
        """
        response = self.sns_client.create_platform_endpoint(
            PlatformApplicationArn=platform_application_arn,
            Token=device_token,
            CustomUserData=custom_user_data,
        )

        return response

    def delete_device_endpoint(self, endpoint_arn: str, **kwargs) -> dict:
        """
        Deletes the platform endpoint for the given endpoint ARN.
        Args:
            endpoint_arn (str): The ARN of the platform endpoint to delete.
        Returns:
            dict: Response from the SNS client operation, which includes the result of
            the delete operation or an error message if the operation fails.
        """
        response = self.sns_client.delete_endpoint(EndpointArn=endpoint_arn)
        return response

    def update_device_endpoint(
        self, endpoint_arn: str, custom_user_data: str = "", **kwargs
    ) -> dict:
        """
        Updates the CustomUserData for the given platform endpoint.
        Args:
            endpoint_arn (str): The ARN of the platform endpoint to update.
            custom_user_data (str): The new custom user data to
            associate with the endpoint.
                This should be a JSON-formatted string representing user-specific data.
        Returns:
            dict: Response from the SNS client operation, which includes the
            updated platform endpoint details or an error message if the operation fails.
        """
        response = self.sns_client.set_endpoint_attributes(
            EndpointArn=endpoint_arn, Attributes={"CustomUserData": custom_user_data}
        )
        return response



import json
from dataclasses import dataclass
from typing import Any, Dict, List, ClassVar

from notihub.notifiers.aws.clients.base_aws_client import BaseAWSClient


@dataclass
class SESClient(BaseAWSClient):
    """
    SESClient

    Class used to generate notifications via AWS SES
    """
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str

    def __post_init__(self):
        self.ses_client = self.initialize_client("ses")

    def create_email_template(
        self, template_name: str, subject: str, text_body: str, html_body: str
    ) -> Dict[str, Any]:
        """
        Creates an email template with the given name

        Args:
            template_name (str): The name of the template
            subject (str): The subject of the template
            text_body (str): The text body of the template
            html_body (str): The HTML body of the template

        Returns:
            dict: Response of the client operation
        """
        return self.ses_client.create_template(
            Template={
                "TemplateName": template_name,
                "SubjectPart": subject,
                "HtmlPart": html_body,
                "TextPart": text_body,
            }
        )

    def update_email_template(
        self, template_name: str, subject: str, text_body: str, html_body: str
    ) -> str:
        """
        Updates an email template with the given name

        Args:
            template_name (str): The name of the template
            subject (str): The subject of the template
            text_body (str): The text body of the template
            html_body (str): The HTML body of the template

        Returns:
            dict: Response of the client operation
        """
        return self.ses_client.update_template(
            Template={
                "TemplateName": template_name,
                "SubjectPart": subject,
                "HtmlPart": html_body,
                "TextPart": text_body,
            }
        )

    def get_email_template(self, template_name: str) -> Dict[str, Any]:
        """
        Gets an email template with the given name

        Args:
            template_name (str): The name of the template

        Returns:
            dict: Response with the template data
        """
        return self.ses_client.get_template(
            TemplateName=template_name,
        )

    def delete_email_template(self, template_name: str) -> Dict[str, Any]:
        """
        Deletes an email template with the given name

        Args:
            template_name (str): The name of the template

        Returns:
            dict: Response of the client operation
        """
        return self.ses_client.delete_template(
            TemplateName=template_name,
        )

    def list_email_templates(self) -> List[Dict[str, Any]]:
        """
        Lists all email templates

        Args:
            template_name (str): The name of the template

        Returns:
            list: List of email templates
        """
        return self.ses_client.list_templates()

    def send_email_notification(
        self,
        *,
        email_data: dict,
        recipients: List[str],
        sender: str,
        template: str,
        cc_emails: List[str] = None,
        bcc_emails: List[str] = None,
        subject: str = None,
        **kwargs,
    ) -> str:
        """

        Sends an email notification to the given emails with a template

        Args:
            email_data (dict): The data to be used in the email template
            recipients (List[str]): The recipients of the email
            sender (str): The sender of the email
            template (str): The name of the email template

        Additional arguments:
            subject (str): The subject of the email (not required if template is provided)
            cc_emails (List[str]): The CC emails of the email
            bcc_emails (List[str]): The BCC emails of the email
            \*args: Additional arguments
            \*\*kwargs: Additional keyword arguments

        Returns:
            dict: Response of the client operation
        """
        if subject:
            template_data = self.get_email_template(template_name=template)
            self.update_email_template(
                template_name=template,
                subject=subject,
                text_body=template_data["Template"].get("TextPart"),
                html_body=template_data["Template"].get("HtmlPart"),
            )

        return self.ses_client.send_templated_email(
            Source=sender,
            Destination={
                "ToAddresses": recipients,
                "CcAddresses": cc_emails or [],
                "BccAddresses": bcc_emails or [],
            },
            Template=template,
            TemplateData=json.dumps(email_data),
            **kwargs,
        )



"""

Base client for notifiers

This module contains the base client for notifiers it is used to access the notifiers
via interface

"""

from typing import Type, Optional

from notihub.notifiers.aws.notifier import AWSNotifier
from notihub.notifiers.twilio.notifier import TwilioNotifier


class NotifierClient:
    """
    Notifier client

    Used as interface to access notifiers
    """

    @staticmethod
    def get_aws_notifier(
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        region_name: Optional[str] = None
    ) -> Type[AWSNotifier]:
        """
        Returns an AWS notifier client
        
        If no credentials are provided, the default ones will be used
        using boto3 default config handling
        """
        return AWSNotifier(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )

    @staticmethod
    def get_twilio_notifier(
        account_sid: Optional[str] = None,
        auth_token: Optional[str] = None,
        twilio_phone_number: Optional[str] = None,
    ) -> Type[TwilioNotifier]:
        """
        Returns a Twilio notifier client.
        """
        return TwilioNotifier(
            account_sid=account_sid,
            auth_token=auth_token,
            twilio_phone_number=twilio_phone_number,
        )



"""

AWS Notifier

This module contains the AWS notifier class which is used to send notifications via AWS
SNS, SES or Pinpoint
"""

from dataclasses import dataclass

from notihub.base_notifier import BaseNotifier
from notihub.notifiers.aws.clients.pinpoint_client import PinpointClient
from notihub.notifiers.aws.clients.ses_client import SESClient
from notihub.notifiers.aws.clients.sns_client import SNSClient


@dataclass
class AWSNotifier(SNSClient, SESClient, PinpointClient, BaseNotifier):
    """
    AWSNotifier

    Centralized class to send notifications via AWS SNS, SES or Pinpoint using
    class inheritance to initialize the clients.
    """

    aws_access_key_id: str = None
    aws_secret_access_key: str = None
    region_name: str = None

    def __post_init__(self):
        """
        Initializes the parent client classes after AWSNotifier's own fields are set.
        This ensures that self.sns_client, self.ses_client, and self.pinpoint_client
        are created by their respective parent classes.
        """
        SNSClient.__post_init__(self)
        SESClient.__post_init__(self)
        PinpointClient.__post_init__(self)
